Package: SpatialML
Type: Package
Title: Spatial Machine Learning
Version: 1.8.2
Date: 2026-07-06
Authors@R: c(
    person("Stamatis", "Kalogirou",
           email = "stamatis.science@gmail.com",
           role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-1949-6248")),
    person("Stefanos", "Georganos",
           email = "stefanos.georganos@kau.se",
           role = "aut"))
Description: Implements a spatial extension of the random forest algorithm
    (Georganos et al. (2019) <doi:10.1080/10106049.2019.1595177>). Provides
    a Geographically Weighted Random Forest regression and a routine to
    find the optimal bandwidth (Georganos and Kalogirou (2022)
    <doi:10.3390/ijgi11090471>). A lightweight cross-validation helper for
    tuning the 'mtry' parameter of a random forest and a generator of
    synthetic spatial test data are also included. The package depends on
    'ranger' as its single random-forest back-end.
License: GPL (>= 2)
Encoding: UTF-8
Language: en-GB
LazyData: true
Depends: R (>= 4.1.0)
Imports: ranger (>= 0.15.1), stats, graphics, utils
Suggests: knitr, markdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://stamatisgeoai.eu/
NeedsCompilation: no
Packaged: 2026-07-06 07:28:11 UTC; Stamatis
Author: Stamatis Kalogirou [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-1949-6248>),
  Stefanos Georganos [aut]
Maintainer: Stamatis Kalogirou <stamatis.science@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-06 09:40:09 UTC

export(grf)
export(grf.bw)
export(random.test.data)
export(rf.mtry.optim)

S3method(predict, grf)

importFrom("ranger", "ranger")
importFrom("stats", "cor", "dist", "formula", "predict", "sd", "setNames",
           "terms", "rnorm", "rpois", "runif")
importFrom("graphics", "abline", "par", "plot")
importFrom("utils", "setTxtProgressBar", "txtProgressBar")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 6,
  fig.height = 4,
  warning  = FALSE,
  message  = FALSE
)

## -----------------------------------------------------------------------------
library(SpatialML)

## -----------------------------------------------------------------------------
set.seed(42)
df <- random.test.data(nrows = 8, ncols = 8, vars.no = 3)
head(df)

## -----------------------------------------------------------------------------
set.seed(1)
mtry.opt <- rf.mtry.optim(dep ~ X1 + X2, dataset = df,
                          cv.method = "oob", plot.it = FALSE,
                          verbose = FALSE)
mtry.opt$best.mtry
mtry.opt$results

## ----results = "hide"---------------------------------------------------------
set.seed(1)
bw.search <- grf.bw(dep ~ X1 + X2, dataset = df, kernel = "adaptive",
                    coords = df[, c("X", "Y")],
                    bw.min = 8, bw.max = 18, step = 2,
                    trees = 200, mtry = mtry.opt$best.mtry,
                    verbose = FALSE)

## -----------------------------------------------------------------------------
bw.search$tested.bandwidths
bw.search$Best.BW

## ----results = "hide"---------------------------------------------------------
set.seed(1)
m <- grf(dep ~ X1 + X2, dframe = df, bw = bw.search$Best.BW,
         kernel = "adaptive", coords = df[, c("X", "Y")],
         ntree = 200, mtry = mtry.opt$best.mtry,
         forests = TRUE, print.results = FALSE, progress = FALSE)

## -----------------------------------------------------------------------------
class(m)
m$LocalModelSummary$l.r.OOB              # local OOB R-squared
summary(m$LGofFit$LM_ResOOB)             # local OOB residuals
head(m$Local.Variable.Importance)        # local importance per observation

## ----fig.alt = "Local importance of predictor X1 across the synthetic 8 x 8 grid."----
imp <- m$Local.Variable.Importance$X1
plot(df$X, df$Y, pch = 19, cex = 2,
     col = grDevices::grey(1 - imp / max(imp)),
     xlab = "X", ylab = "Y",
     main = "Local importance of X1 (darker = more important)")

## -----------------------------------------------------------------------------
new.df <- random.test.data(2, 2, vars.no = 3)
predict(m, new.df, x.var.name = "X", y.var.name = "Y")

## -----------------------------------------------------------------------------
predict(m, new.df, x.var.name = "X", y.var.name = "Y",
        local.w = 0.5, global.w = 0.5)

## ----eval = FALSE-------------------------------------------------------------
# data(Income)
# Coords <- Income[, 1:2]
# 
# # 1. Search the optimal bandwidth (be patient)
# bw <- grf.bw(Income01 ~ UnemrT01 + PrSect01,
#              dataset = Income, kernel = "adaptive",
#              coords = Coords, bw.min = 30, bw.max = 80, step = 5)
# 
# # 2. Fit the final model
# m <- grf(Income01 ~ UnemrT01 + PrSect01, dframe = Income,
#          bw = bw$Best.BW, kernel = "adaptive", coords = Coords)
# 
# # 3. Local R-squared
# m$LocalModelSummary$l.r.OOB
# 
# # 4. Map the residuals
# plot(Coords[, 1], Coords[, 2], pch = 19,
#      col = ifelse(m$LGofFit$LM_ResOOB > 0, "red", "blue"),
#      xlab = "X", ylab = "Y", main = "GRF OOB residuals")

# Tiny fixtures shared by the smoke tests. The data set is intentionally
# small (5 x 5 grid -> 25 observations) and the random forests are fitted
# with only 50 trees on a single thread to keep the whole test suite well
# under CRAN's typical check-time budget.

.ensure_spatialml_loaded <- function() {
  if (exists("random.test.data", mode = "function") &&
      exists("grf", mode = "function")) {
    return(invisible(TRUE))
  }

  candidates <- normalizePath(
    c(getwd(), file.path(getwd(), ".."), file.path(getwd(), "..", "..")),
    mustWork = FALSE
  )
  is_pkg_root <- file.exists(file.path(candidates, "DESCRIPTION")) &
    dir.exists(file.path(candidates, "R"))
  pkg_root <- candidates[is_pkg_root][1]
  if (is.na(pkg_root)) {
    stop("Could not find the SpatialML package root. Run tests from the ",
         "package root or use testthat::test_dir('tests/testthat').",
         call. = FALSE)
  }

  r_files <- list.files(file.path(pkg_root, "R"), pattern = "\\.R$",
                        full.names = TRUE)
  for (r_file in r_files) sys.source(r_file, envir = globalenv())
  invisible(TRUE)
}

.ensure_spatialml_loaded()

.toy_data <- function(seed = 1, nrows = 5, ncols = 5, vars.no = 3) {
  set.seed(seed)
  random.test.data(nrows = nrows, ncols = ncols, vars.no = vars.no)
}

.toy_grf <- function(seed = 1, ..., bw = 8, kernel = "adaptive",
                     forests = TRUE, geo.weighted = TRUE) {
  df <- .toy_data(seed)
  set.seed(seed)
  grf(dep ~ X1 + X2, dframe = df, bw = bw, kernel = kernel,
      coords = df[, c("X", "Y")],
      ntree = 50, mtry = 1, nthreads = 1L,
      forests = forests, geo.weighted = geo.weighted,
      print.results = FALSE, progress = FALSE, ...)
}

test_that("grf returns an object of class 'grf' with the expected slots", {
  m <- .toy_grf()

  expect_s3_class(m, "grf")
  expect_s3_class(m$Global.Model, "ranger")

  # All documented slots must be present.
  expected_slots <- c("Global.Model", "Locations",
                      "Local.Variable.Importance", "LGofFit",
                      "Forests", "LocalModelSummary")
  expect_true(all(expected_slots %in% names(m)))

  # Local diagnostics are populated for every observation.
  expect_equal(nrow(m$LGofFit), 25L)
  expect_equal(nrow(m$Local.Variable.Importance), 25L)
  expect_length(m$Forests, 25L)

  # LGofFit has the documented columns.
  expect_named(m$LGofFit,
               c("y", "LM_yfitOOB", "LM_ResOOB", "LM_yfitPred",
                 "LM_ResPred", "LM_MSE", "LM_Rsq100", "LPerm"))

  # LocalModelSummary contains the four headline statistics.
  expect_true(all(c("l.MSE.OOB", "l.r.OOB",
                    "l.MSE.Pred", "l.r.Pred") %in%
                  names(m$LocalModelSummary)))
})

test_that("grf with forests = FALSE drops the Forests slot", {
  m <- .toy_grf(forests = FALSE)
  expect_null(m$Forests)
  expect_s3_class(m, "grf")
})

test_that("grf with the fixed kernel runs end-to-end", {
  df <- .toy_data()
  set.seed(1)
  m <- grf(dep ~ X1 + X2, dframe = df, bw = 3,
           kernel = "fixed", coords = df[, c("X", "Y")],
           ntree = 50, mtry = 1, nthreads = 1L,
           print.results = FALSE, progress = FALSE)
  expect_s3_class(m, "grf")
})

test_that("grf with geo.weighted = FALSE runs end-to-end", {
  m <- .toy_grf(geo.weighted = FALSE)
  expect_s3_class(m, "grf")
})

test_that("grf handles small weighted windows without noisy OOB warnings", {
  df <- .toy_data()
  set.seed(1)
  expect_warning(
    m <- grf(dep ~ X1 + X2, dframe = df, bw = 4,
             kernel = "adaptive", coords = df[, c("X", "Y")],
             ntree = 20, mtry = 1, nthreads = 1L,
             print.results = FALSE, progress = FALSE),
    NA
  )
  expect_s3_class(m, "grf")
  expect_false(any(!is.finite(m$LGofFit$LM_yfitOOB)))
})

test_that("grf validates inputs", {
  df <- .toy_data()
  coords <- df[, c("X", "Y")]

  expect_error(
    grf(dep ~ X1 + X2, dframe = "not a data frame", bw = 8,
        kernel = "adaptive", coords = coords),
    "data frame"
  )
  expect_error(
    grf(dep ~ X1 + X2, dframe = df, bw = -1,
        kernel = "adaptive", coords = coords),
    "positive numeric"
  )
  expect_error(
    grf(dep ~ X1 + X2, dframe = df, bw = 8,
        kernel = "adaptive", coords = coords[1:10, , drop = FALSE]),
    "same number of rows"
  )
  expect_error(
    grf(dep ~ X1 + X2, dframe = df, bw = 1000,
        kernel = "adaptive", coords = coords),
    "between 2 and"
  )
})

test_that("predict() dispatches to predict.grf and returns a numeric vector", {
  m <- .toy_grf()

  set.seed(2)
  newd <- random.test.data(nrows = 2, ncols = 3, vars.no = 3)
  p <- predict(m, newd, x.var.name = "X", y.var.name = "Y")

  expect_type(p, "double")
  expect_length(p, 6L)
  expect_false(any(is.na(p)))
})

test_that("predict.grf errors when the model was fitted with forests = FALSE", {
  m <- .toy_grf(forests = FALSE)
  set.seed(2)
  newd <- random.test.data(nrows = 2, ncols = 2, vars.no = 3)
  expect_error(
    predict(m, newd, x.var.name = "X", y.var.name = "Y"),
    "forests = TRUE"
  )
})

test_that("predict.grf errors on a missing coordinate column", {
  m <- .toy_grf()
  set.seed(2)
  newd <- random.test.data(nrows = 2, ncols = 2, vars.no = 3)
  expect_error(
    predict(m, newd, x.var.name = "Lon", y.var.name = "Y"),
    "Lon"
  )
})

test_that("predict.grf blends local and global predictions correctly", {
  m <- .toy_grf()
  set.seed(2)
  newd <- random.test.data(nrows = 2, ncols = 2, vars.no = 3)

  p_local  <- predict(m, newd, x.var.name = "X", y.var.name = "Y",
                      local.w = 1, global.w = 0)
  p_global <- predict(m, newd, x.var.name = "X", y.var.name = "Y",
                      local.w = 0, global.w = 1)
  p_blend  <- predict(m, newd, x.var.name = "X", y.var.name = "Y",
                      local.w = 0.5, global.w = 0.5)

  # Linearity of the blend: 0.5 * local + 0.5 * global
  expect_equal(p_blend, 0.5 * p_local + 0.5 * p_global, tolerance = 1e-8)
})

test_that("grf.bw returns the expected list shape", {
  df <- .toy_data()
  set.seed(1)
  # 100 trees and bandwidths >= 10 keep the stochastic local fits stable
  # while still exercising a real bandwidth grid.
  res <- grf.bw(dep ~ X1 + X2, dataset = df, kernel = "adaptive",
                coords = df[, c("X", "Y")],
                bw.min = 10, bw.max = 12, step = 1,
                trees = 100, mtry = 1, nthreads = 1L,
                forests = FALSE, verbose = FALSE)

  expect_named(res, c("tested.bandwidths", "Best.BW"))
  expect_s3_class(res$tested.bandwidths, "data.frame")
  expect_named(res$tested.bandwidths,
               c("Bandwidth", "Local", "Mixed", "Low.Local"))
  expect_equal(nrow(res$tested.bandwidths), 3L)
  expect_true(res$Best.BW %in% res$tested.bandwidths$Bandwidth)
  expect_true(is.numeric(res$Best.BW))
})

test_that("grf.bw validates inputs", {
  df <- .toy_data()
  expect_error(
    grf.bw(dep ~ X1 + X2, dataset = "not a df", kernel = "adaptive",
           coords = df[, c("X", "Y")]),
    "data frame"
  )
  expect_error(
    grf.bw(dep ~ X1 + X2, dataset = df, kernel = "adaptive",
           coords = df[, c("X", "Y")],
           bw.min = 20, bw.max = 10),
    "greater than or equal"
  )
})

test_that("rf.mtry.optim with the OOB strategy returns the expected structure", {
  set.seed(1)
  df <- random.test.data(nrows = 5, ncols = 5, vars.no = 4)

  set.seed(1)
  res <- rf.mtry.optim(dep ~ X1 + X2 + X3, dataset = df,
                       num.trees = 50, num.threads = 1L,
                       cv.method = "oob",
                       plot.it = FALSE, verbose = FALSE)

  expect_named(res, c("best.mtry", "results", "cv.method",
                      "num.trees", "call"))
  expect_true(res$best.mtry >= 1L && res$best.mtry <= 3L)
  expect_s3_class(res$results, "data.frame")
  expect_named(res$results,
               c("mtry", "RMSE", "Rsquared", "SDRMSE", "SDRsq"))
  expect_equal(res$cv.method, "oob")
  expect_equal(res$num.trees, 50)
  # OOB does not perform resampling so the SDs are NA.
  expect_true(all(is.na(res$results$SDRMSE)))
  expect_true(all(is.na(res$results$SDRsq)))
})

test_that("rf.mtry.optim caps max.mtry at the number of predictors", {
  set.seed(1)
  df <- random.test.data(nrows = 5, ncols = 5, vars.no = 3)
  set.seed(1)
  res <- rf.mtry.optim(dep ~ X1 + X2, dataset = df,
                       max.mtry = 99, num.trees = 50,
                       num.threads = 1L, cv.method = "oob",
                       plot.it = FALSE, verbose = FALSE)
  expect_lte(max(res$results$mtry), 2L)
})

test_that("rf.mtry.optim validates inputs", {
  df <- .toy_data()
  expect_error(
    rf.mtry.optim(dep ~ X1 + X2, dataset = "not a df"),
    "data frame"
  )
  expect_error(
    rf.mtry.optim(dep ~ X1 + X2, dataset = df, cv.folds = 1),
    "cv.folds"
  )
  expect_error(
    rf.mtry.optim(dep ~ X1 + X2, dataset = df,
                  min.mtry = 5, max.mtry = 1),
    "greater than or equal"
  )
})

test_that("rf.mtry.optim 'cv' strategy runs (skipped on CRAN for time)", {
  skip_on_cran()

  set.seed(1)
  df <- random.test.data(nrows = 5, ncols = 5, vars.no = 3)

  set.seed(1)
  res <- rf.mtry.optim(dep ~ X1 + X2, dataset = df,
                       num.trees = 50, num.threads = 1L,
                       cv.method = "cv", cv.folds = 3,
                       plot.it = FALSE, verbose = FALSE)
  expect_equal(res$cv.method, "cv")
  expect_true(res$best.mtry %in% c(1L, 2L))
  # CV does report fold-to-fold standard deviations.
  expect_false(any(is.na(res$results$SDRMSE)))
})

# =============================================================================
# Optimal bandwidth selection for the Geographically Weighted Random Forest
#
# Exhaustive grid search: for every candidate bandwidth in [bw.min, bw.max]
# (inclusive, with stride 'step') a full grf() is fitted and its OOB local
# R-squared recorded. Two additional R-squared values are reported as
# diagnostics: one based on a 50/50 average of local and global predictions
# and one based on a 25/75 weighted average.
#
# The search is computationally expensive (one full grf() per bandwidth);
# for big data sets restrict the range with bw.min/bw.max and use a coarse
# 'step', then refine.
# =============================================================================

grf.bw <- function(formula, dataset, kernel = "adaptive", coords,
                   bw.min = NULL, bw.max = NULL, step = 1, trees = 500,
                   mtry = NULL, importance = "impurity", nthreads = 1,
                   forests = FALSE, geo.weighted = TRUE,
                   verbose = TRUE, ...) {

  # ---- Argument validation ----
  if (!is.data.frame(dataset))
    stop("'dataset' must be a data frame.", call. = FALSE)
  kernel <- match.arg(kernel, c("adaptive", "fixed"))
  if (missing(coords) || NCOL(coords) < 2L || NROW(coords) != nrow(dataset))
    stop("'coords' must be a 2-column object with the same number of rows ",
         "as 'dataset'.", call. = FALSE)
  if (!is.numeric(step) || length(step) != 1L || step <= 0)
    stop("'step' must be a positive number.", call. = FALSE)

  Obs <- nrow(dataset)
  f <- stats::formula(formula)
  RNames <- attr(stats::terms(f), "term.labels")
  ModelVarNo <- length(RNames)

  DepVarName <- row.names(attr(stats::terms(f), "factors"))[1]
  Y <- dataset[[DepVarName]]

  # Default search range: 5%-95% of Obs (or at least p+2 / >= 20 neighbours).
  if (is.null(bw.min)) bw.min <- max(round(Obs * 0.05, 0), ModelVarNo + 2L, 20L)
  if (is.null(bw.max)) bw.max <- max(round(Obs * 0.95, 0), ModelVarNo + 2L)
  if (bw.max < bw.min)
    stop("'bw.max' must be greater than or equal to 'bw.min'.", call. = FALSE)
  if (is.null(mtry)) mtry <- max(floor(ModelVarNo / 3), 1L)
  if (mtry > ModelVarNo) {
    if (isTRUE(verbose))
      message("'mtry' (", mtry, ") exceeds the number of predictors (",
              ModelVarNo, "); capping at ", ModelVarNo, ".")
    mtry <- ModelVarNo
  }

  # ---- Empty result table ----
  eval.bw.grf <- data.frame(Bandwidth = numeric(0),
                            Local     = numeric(0),
                            Mixed     = numeric(0),
                            Low.Local = numeric(0),
                            stringsAsFactors = FALSE)

  # ---- Main grid search ----
  count <- 1L
  for (abw in seq(from = bw.min, to = bw.max, by = step)) {

    eval.bw.grf[count, 1] <- abw

    grf.a <- grf(formula, dframe = dataset, bw = abw, kernel = kernel,
                 coords = coords, ntree = trees, mtry = mtry,
                 importance = importance, nthreads = nthreads,
                 forests = FALSE, geo.weighted = geo.weighted,
                 print.results = FALSE, progress = FALSE, ...)

    # Local R-squared (OOB) is the primary criterion.
    eval.bw.grf[count, 2] <- grf.a$LocalModelSummary$l.r.OOB

    if (isTRUE(verbose)) {
      message("Bandwidth: ", abw,
              " - R-squared local: ", round(eval.bw.grf[count, 2], 4))
    }

    # 50/50 blend of local OOB and global in-sample predictions.
    eval.bw.grf[count, 3] <- .cor_r2(
      pred = (grf.a$LGofFit$LM_yfitOOB + grf.a$Global.Model$predictions) / 2,
      obs  = Y)

    # 25/75 blend (mostly the global model, used as a stability check).
    eval.bw.grf[count, 4] <- .cor_r2(
      pred = (grf.a$LGofFit$LM_yfitOOB * 0.25) +
             (grf.a$Global.Model$predictions * 0.75),
      obs  = Y)

    count <- count + 1L
  }

  # ---- Pick the bandwidth that maximises the local R-squared ----
  # which.max() returns integer(0) if every value is NA / NaN, which would
  # leave Best.BW empty. Guard explicitly so callers always receive either
  # a single numeric value or an explicit NA they can test against.
  if (all(!is.finite(eval.bw.grf$Local))) {
    warning("All evaluated bandwidths produced non-finite local R-squared; ",
            "no optimal bandwidth could be selected. Try a wider bandwidth ",
            "range, increase 'trees', or set 'geo.weighted = FALSE'.",
            call. = FALSE)
    best.bw <- NA_real_
  } else {
    best.bw <- eval.bw.grf$Bandwidth[which.max(eval.bw.grf$Local)]
  }
  if (isTRUE(verbose))
    message("Best Bandwidth (Based on the Local Model): ", best.bw)

  list(tested.bandwidths = eval.bw.grf, Best.BW = best.bw)
}

# =============================================================================
# Random spatial test data generator
#
# Produces a data frame with one dependent variable (drawn from N(0, 1) or
# Poisson(7)), 'vars.no - 1' uniformly distributed predictors named
# X1, X2, ..., and optionally a regular nrows-by-ncols grid of (X, Y)
# coordinates.
#
# Mostly intended for the package examples and quick simulation studies.
# Call set.seed() before use if you need reproducible draws.
# =============================================================================

random.test.data <- function(nrows = 10, ncols = 10, vars.no = 3,
                             dep.var.dis = c("normal", "poisson"),
                             xycoords = TRUE) {

  dep.var.dis <- match.arg(dep.var.dis)
  if (!is.numeric(nrows)  || nrows  < 1L)
    stop("'nrows' must be a positive integer.",  call. = FALSE)
  if (!is.numeric(ncols)  || ncols  < 1L)
    stop("'ncols' must be a positive integer.",  call. = FALSE)
  if (!is.numeric(vars.no) || vars.no < 2L)
    stop("'vars.no' must be at least 2 (one dependent + one independent).",
         call. = FALSE)

  obs.no <- as.integer(nrows) * as.integer(ncols)

  # Dependent variable. Note: pre-1.7 the "normal" branch incorrectly used
  # runif(); fixed to rnorm().
  dep <- switch(dep.var.dis,
                normal  = stats::rnorm(obs.no),
                poisson = stats::rpois(obs.no, lambda = 7))

  # Predictor matrix, vectorised. Pre-1.7 this was a per-column loop with
  # an off-by-one bug (`for (i in 1:vars.no - 1)` evaluated to 0:(vars.no-1)).
  n.indep <- as.integer(vars.no) - 1L
  vars <- matrix(stats::runif(obs.no * n.indep),
                 nrow = obs.no, ncol = n.indep)
  colnames(vars) <- paste0("X", seq_len(n.indep))

  if (isTRUE(xycoords)) {
    X <- rep(seq_len(nrows), each = ncols)
    Y <- rep(seq_len(ncols), times = nrows)
    out <- data.frame(dep = dep, vars, X = X, Y = Y)
  } else {
    out <- data.frame(dep = dep, vars)
  }
  out
}

test_that("random.test.data returns the expected shape with coordinates", {
  set.seed(1)
  df <- random.test.data(nrows = 5, ncols = 5, vars.no = 3)
  expect_s3_class(df, "data.frame")
  expect_equal(nrow(df), 25L)
  expect_named(df, c("dep", "X1", "X2", "X", "Y"))
})

test_that("random.test.data without coordinates omits X and Y", {
  set.seed(1)
  df <- random.test.data(nrows = 3, ncols = 3, vars.no = 4,
                         xycoords = FALSE)
  expect_equal(nrow(df), 9L)
  expect_named(df, c("dep", "X1", "X2", "X3"))
})

test_that("random.test.data 'normal' draws come from rnorm", {
  set.seed(1)
  df <- random.test.data(nrows = 20, ncols = 20, vars.no = 2,
                         dep.var.dis = "normal")
  # rnorm has mean approx 0 and routinely produces negative values; the
  # earlier (buggy) version drew from runif and would fail both checks.
  expect_lt(abs(mean(df$dep)), 0.5)
  expect_true(any(df$dep < 0))
})

test_that("random.test.data 'poisson' draws are non-negative integers", {
  set.seed(1)
  df <- random.test.data(nrows = 5, ncols = 5, vars.no = 2,
                         dep.var.dis = "poisson")
  expect_true(all(df$dep >= 0))
  expect_true(all(df$dep == as.integer(df$dep)))
})

test_that("random.test.data validates inputs", {
  expect_error(random.test.data(vars.no = 1L), "vars.no")
  expect_error(random.test.data(nrows = 0L),   "nrows")
  expect_error(random.test.data(ncols = 0L),   "ncols")
  expect_error(random.test.data(dep.var.dis = "uniform"), "should be one of")
})

# =============================================================================
# Predict method for objects of class "grf"
#
# For each row in 'new.data':
#   1. compute the Euclidean distance to every training location stored in
#      object$Locations,
#   2. pick the local Random Forest fitted at the closest training location,
#   3. blend its prediction with the global Random Forest prediction
#      according to the weights local.w and global.w.
#
# 'object$Forests' must be present (i.e. grf() must have been called with
# forests = TRUE), otherwise the local forests are not available and the
# function stops with a clear error message.
# =============================================================================

predict.grf <- function(object, new.data, x.var.name, y.var.name,
                        local.w = 1, global.w = 0, ...) {

  # ---- Validate inputs ----
  if (!inherits(object, "grf"))
    stop("'object' must be of class 'grf'.", call. = FALSE)
  if (is.null(object$Forests))
    stop("'object' does not contain local forests; rerun grf() with ",
         "forests = TRUE.", call. = FALSE)
  if (!is.data.frame(new.data))
    stop("'new.data' must be a data frame.", call. = FALSE)
  if (!(x.var.name %in% names(new.data)))
    stop("Column '", x.var.name, "' not found in 'new.data'.", call. = FALSE)
  if (!(y.var.name %in% names(new.data)))
    stop("Column '", y.var.name, "' not found in 'new.data'.", call. = FALSE)
  if (!is.numeric(local.w) || !is.numeric(global.w))
    stop("'local.w' and 'global.w' must be numeric.", call. = FALSE)

  Obs <- nrow(new.data)
  predictions <- numeric(Obs)

  # Coerce coordinates once outside the loop. as.matrix() copes with
  # data.frame, matrix and tibble inputs.
  locations <- as.matrix(object$Locations)

  for (i in seq_len(Obs)) {

    x <- new.data[[x.var.name]][i]
    y <- new.data[[y.var.name]][i]

    # Euclidean distance from the new point to every training location.
    D <- sqrt((x - locations[, 1])^2 + (y - locations[, 2])^2)
    local.model.ID <- which.min(D)

    g.predict <- stats::predict(object$Global.Model,
                                new.data[i, , drop = FALSE], ...)
    g.prediction <- g.predict$predictions

    l.predict <- stats::predict(object$Forests[[local.model.ID]],
                                new.data[i, , drop = FALSE], ...)
    l.prediction <- l.predict$predictions

    # Linear blend; with the defaults (local.w = 1, global.w = 0) the result
    # is the local prediction only.
    predictions[i] <- global.w * g.prediction[1] + local.w * l.prediction[1]
  }

  predictions
}

# Internal helpers (not exported).

# Squared correlation between predictions and observations, equivalent to
# the "Rsquared" element of caret::postResample(). Pairwise-complete to
# tolerate NA. Returns NA if fewer than two paired observations are usable.
.cor_r2 <- function(pred, obs) {
  ok <- !is.na(pred) & !is.na(obs)
  if (sum(ok) < 2L) return(NA_real_)
  r <- suppressWarnings(stats::cor(pred[ok], obs[ok]))
  if (is.na(r)) NA_real_ else r * r
}

# Root mean squared error.
.rmse <- function(pred, obs) {
  ok <- !is.na(pred) & !is.na(obs)
  if (!any(ok)) return(NA_real_)
  sqrt(mean((pred[ok] - obs[ok])^2))
}

# =============================================================================
# Optimal mtry selection for a Random Forest
#
# Lightweight replacement for the previous caret::train(method = "rf")
# implementation. Three evaluation strategies are provided:
#
#   * "oob"        - one ranger fit per mtry; OOB error is the criterion.
#                    Fast, no resampling overhead, statistically valid for
#                    Random Forests, and the default.
#   * "cv"         - manual k-fold cross-validation (one repeat).
#   * "repeatedcv" - manual repeated k-fold CV (cv.repeats * cv.folds fits
#                    per mtry).
#
# For all three strategies the criterion to *minimise* is the average RMSE
# across folds (or the OOB RMSE in the "oob" case). Average R-squared is
# also recorded for diagnostics.
# =============================================================================

rf.mtry.optim <- function(formula, dataset,
                          min.mtry = NULL, max.mtry = NULL,
                          mtry.step = 1, num.trees = 500,
                          cv.method = c("oob", "repeatedcv", "cv"),
                          cv.folds = 10, cv.repeats = 5,
                          num.threads = NULL,
                          plot.it = TRUE, verbose = TRUE, ...) {

  # ---- Argument validation ----
  cv.method <- match.arg(cv.method)
  if (!is.data.frame(dataset))
    stop("'dataset' must be a data frame.", call. = FALSE)
  if (!is.numeric(mtry.step) || mtry.step <= 0)
    stop("'mtry.step' must be a positive number.", call. = FALSE)
  if (!is.numeric(num.trees) || num.trees < 1L)
    stop("'num.trees' must be a positive integer.", call. = FALSE)
  if (!is.numeric(cv.folds) || cv.folds < 2L)
    stop("'cv.folds' must be at least 2.", call. = FALSE)
  if (!is.numeric(cv.repeats) || cv.repeats < 1L)
    stop("'cv.repeats' must be at least 1.", call. = FALSE)

  # ---- Build the candidate mtry grid ----
  f <- stats::formula(formula)
  RNames <- attr(stats::terms(f), "term.labels")
  ModelVarNo <- length(RNames)

  if (is.null(min.mtry)) min.mtry <- 1L
  if (is.null(max.mtry)) max.mtry <- ModelVarNo
  if (max.mtry < min.mtry)
    stop("'max.mtry' must be greater than or equal to 'min.mtry'.",
         call. = FALSE)
  if (max.mtry > ModelVarNo) {
    if (verbose)
      message("'max.mtry' (", max.mtry, ") exceeds the number of predictors ",
              "(", ModelVarNo, "); capping at ", ModelVarNo, ".")
    max.mtry <- ModelVarNo
  }

  mtry.grid <- seq.int(from = as.integer(min.mtry),
                       to   = as.integer(max.mtry),
                       by   = as.integer(mtry.step))

  # Response, used only for CV-based RMSE / R-squared.
  DepVarName <- row.names(attr(stats::terms(f), "factors"))[1]
  Y <- dataset[[DepVarName]]
  N <- nrow(dataset)

  # ---- Evaluate each mtry ----
  results <- data.frame(mtry     = mtry.grid,
                        RMSE     = NA_real_,
                        Rsquared = NA_real_,
                        SDRMSE   = NA_real_,
                        SDRsq    = NA_real_)

  for (i in seq_along(mtry.grid)) {
    m <- mtry.grid[i]

    if (cv.method == "oob") {

      # Single ranger fit; ranger reports OOB MSE and OOB R-squared directly.
      fit <- ranger::ranger(formula = f, data = dataset,
                            num.trees = num.trees, mtry = m,
                            num.threads = num.threads, ...)
      results$RMSE[i]     <- sqrt(fit$prediction.error)
      results$Rsquared[i] <- fit$r.squared

    } else {

      # Manual (repeated) k-fold cross-validation. "cv" is just the special
      # case cv.repeats = 1 to keep the code path single.
      reps <- if (cv.method == "cv") 1L else as.integer(cv.repeats)

      rmse_runs <- numeric(reps * cv.folds)
      r2_runs   <- numeric(reps * cv.folds)
      k <- 1L

      for (rep_i in seq_len(reps)) {
        # Stratification is not needed for a continuous response; balanced
        # fold sizes are obtained by recycling 1:cv.folds across N rows.
        fold.id <- sample(rep(seq_len(cv.folds), length.out = N))

        for (fold in seq_len(cv.folds)) {
          test.idx  <- which(fold.id == fold)
          train.idx <- setdiff(seq_len(N), test.idx)

          fit <- ranger::ranger(formula = f,
                                data = dataset[train.idx, , drop = FALSE],
                                num.trees = num.trees, mtry = m,
                                num.threads = num.threads, ...)
          pred <- stats::predict(fit, dataset[test.idx, , drop = FALSE],
                                 num.threads = num.threads)$predictions

          rmse_runs[k] <- .rmse(pred, Y[test.idx])
          r2_runs[k]   <- .cor_r2(pred, Y[test.idx])
          k <- k + 1L
        }
      }

      results$RMSE[i]     <- mean(rmse_runs)
      results$Rsquared[i] <- mean(r2_runs, na.rm = TRUE)
      results$SDRMSE[i]   <- stats::sd(rmse_runs)
      results$SDRsq[i]    <- stats::sd(r2_runs, na.rm = TRUE)
    }

    if (verbose)
      message("mtry = ", m,
              " | RMSE = ", round(results$RMSE[i], 4),
              " | R-squared = ", round(results$Rsquared[i], 4))
  }

  # ---- Best mtry by lowest RMSE ----
  best.idx  <- which.min(results$RMSE)
  best.mtry <- results$mtry[best.idx]
  if (verbose)
    message("Best mtry: ", best.mtry,
            " (RMSE = ", round(results$RMSE[best.idx], 4), ")")

  # ---- Optional plot ----
  if (isTRUE(plot.it)) {
    op <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(op), add = TRUE)
    graphics::plot(results$mtry, results$RMSE, type = "b", pch = 19,
                   xlab = "mtry", ylab = "RMSE",
                   main = paste0("rf.mtry.optim (", cv.method, ")"))
    graphics::abline(v = best.mtry, lty = 2, col = "red")
  }

  list(best.mtry = best.mtry,
       results   = results,
       cv.method = cv.method,
       num.trees = num.trees,
       call      = match.call())
}

# =============================================================================
# Geographically Weighted Random Forest (GRF)
#
# For each observation i in space the function fits a Random Forest on the
# local neighbourhood of i (defined by the k nearest neighbours when the
# kernel is "adaptive", or by all observations within a fixed radius bw when
# the kernel is "fixed"). When geo.weighted = TRUE local observations are
# weighted by the bi-square kernel  w_ij = (1 - (d_ij / h)^2)^2,  where
# d_ij is the Euclidean distance between i and j and h is the local
# bandwidth (the largest distance retained in the neighbourhood).
#
# References:
#   Georganos et al. (2019) Geocarto International, doi:10.1080/10106049.2019.1595177
#   Georganos & Kalogirou (2022) ISPRS IJGI 11(9):471, doi:10.3390/ijgi11090471
# =============================================================================

grf <- function(formula, dframe, bw, kernel = c("adaptive", "fixed"),
                coords, ntree = 500, mtry = NULL, importance = "impurity",
                nthreads = NULL, forests = TRUE, geo.weighted = TRUE,
                print.results = TRUE, progress = print.results,
                oob.fallback = c("loo", "inbag"), ...) {

  # ---------------------------------------------------------------------------
  # 1. Argument validation
  # ---------------------------------------------------------------------------
  kernel <- match.arg(kernel)
  oob.fallback <- match.arg(oob.fallback)

  if (!is.data.frame(dframe))
    stop("'dframe' must be a data frame.", call. = FALSE)
  if (missing(coords) || is.null(coords))
    stop("'coords' must be supplied as a 2-column matrix or data frame.",
         call. = FALSE)
  if (NCOL(coords) < 2L)
    stop("'coords' must have at least two columns (X, Y).", call. = FALSE)
  if (NROW(coords) != nrow(dframe))
    stop("'coords' and 'dframe' must have the same number of rows.",
         call. = FALSE)
  if (missing(bw) || !is.numeric(bw) || length(bw) != 1L || bw <= 0)
    stop("'bw' must be a single positive numeric value.", call. = FALSE)
  if (!is.numeric(ntree) || ntree < 1L)
    stop("'ntree' must be a positive integer.", call. = FALSE)

  start.time <- Sys.time()

  # ---------------------------------------------------------------------------
  # 2. Parse the formula and pull the response from the data frame
  # ---------------------------------------------------------------------------
  f <- stats::formula(formula)
  RNames <- attr(stats::terms(f), "term.labels")
  ModelVarNo <- length(RNames)
  if (ModelVarNo < 1L)
    stop("The model formula must contain at least one predictor.",
         call. = FALSE)

  DepVarName <- row.names(attr(stats::terms(f), "factors"))[1]
  if (!(DepVarName %in% names(dframe)))
    stop("Dependent variable '", DepVarName, "' not found in 'dframe'.",
         call. = FALSE)
  Y <- dframe[[DepVarName]]

  K   <- ModelVarNo + 1L            # parameters used in AIC formula
  Obs <- nrow(dframe)

  # Default mtry follows the textbook regression heuristic floor(p / 3).
  if (is.null(mtry)) mtry <- max(floor(ModelVarNo / 3), 1L)

  # For an adaptive kernel, validate the neighbour count up front.
  Ne <- NA_integer_
  if (kernel == "adaptive") {
    if (bw < 2L || bw > Obs)
      stop("For an adaptive kernel, 'bw' (number of neighbours) must be ",
           "between 2 and the number of observations.", call. = FALSE)
    Ne <- as.integer(round(bw))
  }

  if (print.results) {
    message("\nNumber of Observations: ", Obs)
    message("Number of Independent Variables: ", ModelVarNo)
    if (kernel == "adaptive") {
      message("Kernel: Adaptive\nNeighbours: ", Ne)
    } else {
      message("Kernel: Fixed\nBandwidth: ", bw)
    }
  }

  # ---------------------------------------------------------------------------
  # 3. Global random forest (used as benchmark and for blended predictions)
  # ---------------------------------------------------------------------------
  Gl.Model <- ranger::ranger(formula = f, data = dframe, num.trees = ntree,
                             mtry = mtry, importance = importance,
                             num.threads = nthreads, ...)

  # In-sample predictions of the global model. These are *not* OOB; they are
  # used only for the optional blended prediction in predict.grf().
  Predict <- stats::predict(Gl.Model, dframe, num.threads = nthreads)
  yhat <- Predict$predictions

  if (print.results) {
    message("\n--------------- Global ML Model Summary ---------------\n")
    print(Gl.Model)

    message("\nImportance:\n")
    print(Gl.Model$variable.importance)

    g.RSS  <- sum((Y - yhat)^2)
    g.TSS  <- sum((Y - mean(Y))^2)
    g.r    <- 1 - (g.RSS / g.TSS)
    g.AIC  <- 2 * K + Obs * log(g.RSS / Obs)
    g.AICc <- g.AIC + ((2 * K * (K + 1)) / (Obs - K - 1))

    message("\nMean Square Error (Not OOB): ", round(g.RSS / Obs, 3))
    message("R-squared (Not OOB) %: ",        round(100 * g.r, 3))
    message("AIC (Not OOB): ",                round(g.AIC, 3))
    message("AICc (Not OOB): ",               round(g.AICc, 3))
  }

  # ---------------------------------------------------------------------------
  # 4. Pre-compute the full pairwise distance matrix
  #
  # Memory note: this is an N x N matrix of doubles (8 * N^2 bytes). For
  # N around 5000 this already costs 200 MB; very large datasets may need
  # an external nearest-neighbour search (e.g. RANN, FNN), but staying in
  # base R keeps the dependency footprint minimal.
  # ---------------------------------------------------------------------------
  Dij <- as.matrix(stats::dist(coords))

  # ---------------------------------------------------------------------------
  # 5. Allocate the result containers
  # ---------------------------------------------------------------------------
  LM_Forests <- if (isTRUE(forests)) vector("list", Obs) else NULL

  LM_LEst <- as.data.frame(stats::setNames(
    replicate(ModelVarNo, numeric(Obs), simplify = FALSE), RNames))

  LM_GofFit <- data.frame(y           = numeric(Obs),
                          LM_yfitOOB  = numeric(Obs),
                          LM_ResOOB   = numeric(Obs),
                          LM_yfitPred = numeric(Obs),
                          LM_ResPred  = numeric(Obs),
                          LM_MSE      = numeric(Obs),
                          LM_Rsq100   = numeric(Obs),
                          LPerm       = numeric(Obs))

  # Optional progress bar; only useful in interactive sessions.
  show.pb <- isTRUE(progress) && interactive() && Obs > 1L
  if (show.pb) {
    pb <- utils::txtProgressBar(min = 0, max = Obs, style = 3)
    on.exit(close(pb), add = TRUE)
  }

  # ---------------------------------------------------------------------------
  # 6. Local random forests
  # ---------------------------------------------------------------------------
  for (m in seq_len(Obs)) {

    # 6a. Identify the local neighbourhood by index, *without* materialising
    # a per-iteration copy of dframe. The original implementation built
    # data.frame(dframe, DNeighbour = ...) on every iteration, which copied
    # the entire data frame N times.
    d   <- Dij[, m]
    # Keep the focal row first. This is important when two or more rows share
    # identical coordinates: ranger's training predictions are indexed in the
    # row order of SubSet, and the first prediction is treated as the focal one.
    ord <- c(m, setdiff(order(d), m))

    if (kernel == "adaptive") {
      keep <- ord[seq_len(Ne)]
      Kernel_H <- d[keep[Ne]]      # largest distance retained -> bandwidth
    } else {
      keep <- ord[d[ord] <= bw]
      Kernel_H <- bw
      if (length(keep) < (ModelVarNo + 1L))
        stop("Not enough neighbours within the fixed bandwidth at observation ",
             m, ". Increase 'bw'.", call. = FALSE)
    }

    SubSet <- dframe[keep, , drop = FALSE]

    # 6b. Bi-square geographical weights, zero outside the bandwidth. If all
    # selected observations have the same coordinates as the focal one, the
    # bandwidth is zero and the kernel is undefined; equal weights are the
    # least surprising fallback for that degenerate geometry.
    if (!is.finite(Kernel_H) || Kernel_H <= 0) {
      Wts <- rep(1, length(keep))
    } else {
      Wts <- (1 - (d[keep] / Kernel_H)^2)^2
      Wts[!is.finite(Wts)] <- 0
    }

    # 6c. Fit the local forest. Wrapped in a closure so the weighted and
    # unweighted paths share argument plumbing.
    fit_local <- function(weights = NULL) {
      ranger::ranger(formula = f, data = SubSet, num.trees = ntree,
                     mtry = mtry, importance = importance,
                     case.weights = weights,
                     num.threads = nthreads, ...)
    }

    if (isTRUE(geo.weighted)) {
      Lcl.Model <- fit_local(Wts)
    } else {
      Lcl.Model <- fit_local(NULL)
    }
    counter <- 1L

    if (isTRUE(forests)) LM_Forests[[m]] <- Lcl.Model

    # 6d. Record per-observation diagnostics. Indexing the importance vector
    # by name (rather than by position) keeps the result correct even if
    # ranger ever changes the order of variable.importance.
    local.importance <- Lcl.Model$variable.importance
    if (is.null(local.importance) || length(local.importance) == 0L) {
      local.importance <- stats::setNames(rep(NA_real_, ModelVarNo), RNames)
    }
    LM_LEst[m, ] <- local.importance[RNames]

    # In-sample prediction of the focal observation (re-used as the OOB
    # fallback below).
    l.predict <- stats::predict(Lcl.Model, dframe[m, , drop = FALSE],
                                num.threads = nthreads)
    oob.fit  <- Lcl.Model$predictions[[1]]
    pred.fit <- l.predict$predictions

    # Defensive OOB fallback. ranger returns NaN when no tree leaves the
    # focal row out-of-bag, which can happen with very small local windows or
    # highly concentrated case weights. Instead of retrying and printing a
    # warning, the default fits one extra local model without the focal row
    # and predicts it out-of-sample. If that is not possible, use the
    # in-sample prediction as a last resort so summary statistics remain
    # finite.
    if (length(oob.fit) != 1L || !is.finite(oob.fit)) {
      counter <- 2L
      loo.fit <- NA_real_
      if (oob.fallback == "loo" && nrow(SubSet) > 1L) {
        loo.weights <- if (isTRUE(geo.weighted)) Wts[-1] else NULL
        if (!is.null(loo.weights)) {
          loo.weights[!is.finite(loo.weights)] <- 0
          if (!any(loo.weights > 0)) loo.weights <- NULL
        }
        loo.fit <- tryCatch({
          loo.model <- ranger::ranger(
            formula = f, data = SubSet[-1, , drop = FALSE],
            num.trees = ntree, mtry = mtry, importance = importance,
            case.weights = loo.weights, num.threads = nthreads, ...)
          stats::predict(loo.model, SubSet[1, , drop = FALSE],
                         num.threads = nthreads)$predictions[1]
        }, error = function(e) NA_real_)
      }
      oob.fit <- if (length(loo.fit) == 1L && is.finite(loo.fit)) {
        loo.fit
      } else {
        pred.fit
      }
    }

    LM_GofFit[m, "y"]           <- Y[m]
    LM_GofFit[m, "LM_yfitOOB"]  <- oob.fit
    LM_GofFit[m, "LM_ResOOB"]   <- LM_GofFit[m, "y"] - oob.fit
    LM_GofFit[m, "LM_yfitPred"] <- pred.fit
    LM_GofFit[m, "LM_ResPred"]  <- LM_GofFit[m, "y"] - pred.fit
    LM_GofFit[m, "LM_MSE"]      <- Lcl.Model$prediction.error
    LM_GofFit[m, "LM_Rsq100"]   <- Lcl.Model$r.squared
    LM_GofFit[m, "LPerm"]       <- counter

    if (show.pb) utils::setTxtProgressBar(pb, m)
  }

  # ---------------------------------------------------------------------------
  # 7. Assemble the output object
  # ---------------------------------------------------------------------------
  grf.out <- list(Global.Model              = Gl.Model,
                  Locations                 = coords,
                  Local.Variable.Importance = LM_LEst,
                  LGofFit                   = LM_GofFit)
  if (isTRUE(forests)) grf.out$Forests <- LM_Forests

  # Local variable-importance summary statistics
  lvi <- data.frame(
    Min  = apply(grf.out$Local.Variable.Importance, 2, min),
    Max  = apply(grf.out$Local.Variable.Importance, 2, max),
    Mean = apply(grf.out$Local.Variable.Importance, 2, mean),
    StD  = apply(grf.out$Local.Variable.Importance, 2, stats::sd))

  # Aggregate goodness-of-fit
  l.RSS.OOB  <- sum(grf.out$LGofFit$LM_ResOOB^2)
  l.RSS.Pred <- sum(grf.out$LGofFit$LM_ResPred^2)
  TSS        <- sum((grf.out$LGofFit$y - mean(grf.out$LGofFit$y))^2)

  l.r.OOB     <- 1 - (l.RSS.OOB  / TSS)
  l.AIC.OOB   <- 2 * K + Obs * log(l.RSS.OOB  / Obs)
  l.AICc.OOB  <- l.AIC.OOB  + ((2 * K * (K + 1)) / (Obs - K - 1))

  l.r.Pred    <- 1 - (l.RSS.Pred / TSS)
  l.AIC.Pred  <- 2 * K + Obs * log(l.RSS.Pred / Obs)
  l.AICc.Pred <- l.AIC.Pred + ((2 * K * (K + 1)) / (Obs - K - 1))

  if (print.results) {
    message("\n--------------- Local Model Summary ---------------\n")
    message("\nResiduals OOB:\n")
    print(summary(grf.out$LGofFit$LM_ResOOB))

    message("\nResiduals Predicted (Not OOB):\n")
    print(summary(grf.out$LGofFit$LM_ResPred))

    message("\nLocal Variable Importance:\n")
    print(lvi)

    message("\nMean squared error (OOB): ",       round(l.RSS.OOB / Obs, 3))
    message("R-squared (OOB) %: ",                round(100 * l.r.OOB, 3))
    message("AIC (OOB): ",                        round(l.AIC.OOB, 3))
    message("AICc (OOB): ",                       round(l.AICc.OOB, 3))
    message("Mean squared error Predicted (Not OOB): ",
            round(l.RSS.Pred / Obs, 3))
    message("R-squared Predicted (Not OOB) %: ",  round(100 * l.r.Pred, 3))
    message("AIC Predicted (Not OOB): ",          round(l.AIC.Pred, 3))
    message("AICc Predicted (Not OOB): ",         round(l.AICc.Pred, 3))
  }

  grf.out$LocalModelSummary <- list(
    l.VariableImportance = lvi,
    l.MSE.OOB            = l.RSS.OOB  / Obs,
    l.r.OOB              = l.r.OOB,
    l.MSE.Pred           = l.RSS.Pred / Obs,
    l.r.Pred             = l.r.Pred,
    l.AIC.OOB            = l.AIC.OOB,
    l.AICc.OOB           = l.AICc.OOB,
    l.AIC.Pred           = l.AIC.Pred,
    l.AICc.Pred          = l.AICc.Pred
  )

  end.time <- Sys.time()
  time.taken <- as.numeric(end.time - start.time, units = "secs")
  if (print.results)
    message("\nCalculation time (in seconds): ", round(time.taken, 4))

  # The S3 class is what makes predict(grf_obj, ...) dispatch to predict.grf().
  class(grf.out) <- "grf"
  grf.out
}

---
title: "Geographically Weighted Random Forest with SpatialML"
author: "Stamatis Kalogirou and Stefanos Georganos"
date: "`r format(Sys.Date(), '%B %d, %Y')`"
output: markdown::html_format
vignette: >
  %\VignetteIndexEntry{Geographically Weighted Random Forest with SpatialML}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 6,
  fig.height = 4,
  warning  = FALSE,
  message  = FALSE
)
```

## 1. Introduction

**Geographically Weighted Random Forest (GRF)** is a spatial extension of
the Random Forest algorithm. It was introduced by Georganos *et al.*
(2019) and refined in Georganos and Kalogirou (2022). For each
observation *i* the algorithm fits a *local* Random Forest using only
the observations that fall in the neighbourhood of *i* (defined by the
*k* nearest neighbours when the kernel is *adaptive*, or by all the
observations within a distance *bw* when the kernel is *fixed*). When
`geo.weighted = TRUE`, observations within the neighbourhood are
weighted by the bi-square kernel
$$
w_{ij} \;=\; \left(1 - (d_{ij}/h)^{2}\right)^{2},
$$
where $d_{ij}$ is the Euclidean distance between observations *i* and
*j* and $h$ is the largest distance retained in the neighbourhood. The
final model is the collection of all local random forests plus a global
random forest fitted on the whole sample.

The package exports four user-facing functions:

| Function            | Purpose                                                  |
|---------------------|----------------------------------------------------------|
| `grf()`             | Fit a Geographically Weighted Random Forest              |
| `grf.bw()`          | Search the optimal bandwidth                             |
| `predict.grf()`     | Predict at new spatial locations (via S3 `predict()`)    |
| `rf.mtry.optim()`   | Tune the global `mtry` parameter (OOB or k-fold CV)      |
| `random.test.data()`| Generate small synthetic spatial data for testing        |

The package uses **ranger** as its random-forest back-end. Undefined local
out-of-bag predictions are handled with a quiet leave-one-out fallback.

```{r}
library(SpatialML)
```

## 2. Quick start with synthetic data

We start with a small synthetic dataset created on a regular 8 x 8 grid.
`random.test.data()` returns one dependent variable (`dep`), two random
predictors (`X1`, `X2`) and the grid coordinates (`X`, `Y`).

```{r}
set.seed(42)
df <- random.test.data(nrows = 8, ncols = 8, vars.no = 3)
head(df)
```

### 2.1 Tuning `mtry` globally

Before fitting the GRF we tune the `mtry` parameter on the **global**
data. Out-of-bag error (the default) is fast and statistically valid for
Random Forests.

```{r}
set.seed(1)
mtry.opt <- rf.mtry.optim(dep ~ X1 + X2, dataset = df,
                          cv.method = "oob", plot.it = FALSE,
                          verbose = FALSE)
mtry.opt$best.mtry
mtry.opt$results
```

Use `cv.method = "repeatedcv"` for a more rigorous evaluation when the
sample is small.

### 2.2 Optimal bandwidth

`grf.bw()` evaluates a grid of candidate bandwidths and returns the one
that maximises the local OOB R-squared.

```{r, results = "hide"}
set.seed(1)
bw.search <- grf.bw(dep ~ X1 + X2, dataset = df, kernel = "adaptive",
                    coords = df[, c("X", "Y")],
                    bw.min = 8, bw.max = 18, step = 2,
                    trees = 200, mtry = mtry.opt$best.mtry,
                    verbose = FALSE)
```

```{r}
bw.search$tested.bandwidths
bw.search$Best.BW
```

### 2.3 Fitting the GRF

With both `mtry` and the bandwidth chosen we fit the final model. The
`forests = TRUE` argument is required if you want to call `predict()`
later on new data.

```{r, results = "hide"}
set.seed(1)
m <- grf(dep ~ X1 + X2, dframe = df, bw = bw.search$Best.BW,
         kernel = "adaptive", coords = df[, c("X", "Y")],
         ntree = 200, mtry = mtry.opt$best.mtry,
         forests = TRUE, print.results = FALSE, progress = FALSE)
```

### 2.4 Inspecting the fit

The fitted object is an S3 object of class `"grf"`. Useful slots:

```{r}
class(m)
m$LocalModelSummary$l.r.OOB              # local OOB R-squared
summary(m$LGofFit$LM_ResOOB)             # local OOB residuals
head(m$Local.Variable.Importance)        # local importance per observation
```

The columns of `m$Local.Variable.Importance` are the predictors and the
rows correspond, in order, to the observations of `dframe`. To explore
the spatial pattern you can map them with any plotting framework you
like:

```{r, fig.alt = "Local importance of predictor X1 across the synthetic 8 x 8 grid."}
imp <- m$Local.Variable.Importance$X1
plot(df$X, df$Y, pch = 19, cex = 2,
     col = grDevices::grey(1 - imp / max(imp)),
     xlab = "X", ylab = "Y",
     main = "Local importance of X1 (darker = more important)")
```

### 2.5 Predicting at new locations

`predict()` dispatches to `predict.grf()` because `m` has class
`"grf"`. For each new observation the local random forest fitted at the
geographically nearest training location is used.

```{r}
new.df <- random.test.data(2, 2, vars.no = 3)
predict(m, new.df, x.var.name = "X", y.var.name = "Y")
```

By default the global random forest receives weight zero
(`global.w = 0`). Setting `global.w` and `local.w` to non-zero values
returns a linear blend of the two predictions and is a useful sensitivity
test when the local model is noisy.

```{r}
predict(m, new.df, x.var.name = "X", y.var.name = "Y",
        local.w = 0.5, global.w = 0.5)
```

## 3. Real-world example: Greek municipal income

The `Income` dataset (shipped with the package) contains 325
municipalities of Greece with their centroid coordinates and four
socio-economic variables. Fitting a full GRF on this data is heavier
than the toy example above; the chunk below is therefore not evaluated
inside the vignette but copy-paste it into an interactive session.

```{r, eval = FALSE}
data(Income)
Coords <- Income[, 1:2]

# 1. Search the optimal bandwidth (be patient)
bw <- grf.bw(Income01 ~ UnemrT01 + PrSect01,
             dataset = Income, kernel = "adaptive",
             coords = Coords, bw.min = 30, bw.max = 80, step = 5)

# 2. Fit the final model
m <- grf(Income01 ~ UnemrT01 + PrSect01, dframe = Income,
         bw = bw$Best.BW, kernel = "adaptive", coords = Coords)

# 3. Local R-squared
m$LocalModelSummary$l.r.OOB

# 4. Map the residuals
plot(Coords[, 1], Coords[, 2], pch = 19,
     col = ifelse(m$LGofFit$LM_ResOOB > 0, "red", "blue"),
     xlab = "X", ylab = "Y", main = "GRF OOB residuals")
```

## 4. Practical tips

- **Bandwidth.** A *small* bandwidth privileges local detail (and risks
  high variance); a *large* bandwidth approaches the global Random
  Forest. Plot the columns of `bw.search$tested.bandwidths` to inspect
  the trade-off.
- **`forests = FALSE`.** If you only need diagnostics and not prediction
  on new points, set `forests = FALSE`. The output object is then much
  smaller (only one ranger object is kept, the global one).
- **`geo.weighted = FALSE`.** Disables the bi-square case weights but
  keeps the local sub-sampling. Useful for ablation studies.
- **Reproducibility.** All randomness is delegated to `ranger`. Call
  `set.seed()` before any of the package functions (the package itself
  does not call `set.seed()`).

For tutorials, related publications and contact details visit the
maintainer's website at <https://stamatisgeoai.eu/>.

## References

Georganos, S., Grippa, T., Niang Gadiaga, A., Linard, C., Lennert, M.,
Vanhuysse, S., Mboga, N., Wolff, E. and Kalogirou, S. (2019)
Geographical Random Forests: A Spatial Extension of the Random Forest
Algorithm to Address Spatial Heterogeneity in Remote Sensing and
Population Modelling. *Geocarto International*. DOI:
[10.1080/10106049.2019.1595177](https://doi.org/10.1080/10106049.2019.1595177).

Georganos, S. and Kalogirou, S. (2022) A Forest of Forests: A Spatially
Weighted and Computationally Efficient Formulation of Geographical
Random Forests. *ISPRS International Journal of Geo-Information*,
11(9), 471. DOI:
[10.3390/ijgi11090471](https://doi.org/10.3390/ijgi11090471).

Wright, M. N. and Ziegler, A. (2017) ranger: A Fast Implementation of
Random Forests for High Dimensional Data in C++ and R. *Journal of
Statistical Software*, 77(1), 1-17. DOI:
[10.18637/jss.v077.i01](https://doi.org/10.18637/jss.v077.i01).

\name{predict.grf}
\alias{predict.grf}
\title{Predict Method for a Geographically Weighted Random Forest}
\description{
Generates predictions for new observations using a fitted
\code{\link{grf}} object. For each new observation the local random
forest fitted at the geographically nearest training location is used.
The user can blend local and global predictions through the
\code{local.w} and \code{global.w} weights.
}
\usage{
\method{predict}{grf}(object, new.data, x.var.name, y.var.name,
                      local.w = 1, global.w = 0, ...)
}
\arguments{
  \item{object}{an object of class \code{"grf"} created by
    \code{\link{grf}} with \code{forests = TRUE}.}
  \item{new.data}{a data frame containing the predictors required by the
    model formula and the X and Y coordinates of the new observations.}
  \item{x.var.name}{the name of the column in \code{new.data} that
    contains the X coordinate.}
  \item{y.var.name}{the name of the column in \code{new.data} that
    contains the Y coordinate.}
  \item{local.w}{numeric weight applied to the local model prediction.
    Default is \code{1}.}
  \item{global.w}{numeric weight applied to the global model prediction.
    Default is \code{0}.}
  \item{\dots}{further arguments passed to the \code{ranger}
    \code{\link[stats]{predict}} method (e.g. \code{num.threads}).}
}
\details{
For every row \code{i} in \code{new.data} the function computes the
Euclidean distance between
\code{(new.data[i, x.var.name], new.data[i, y.var.name])} and every
training location in \code{object$Locations}, picks the local random
forest fitted at the nearest training location, and combines its
prediction with the prediction of the global random forest using the
weights \code{local.w} and \code{global.w}.
}
\value{
A numeric vector of predictions of length \code{nrow(new.data)}.
}
\references{
Georganos, S., Grippa, T., Niang Gadiaga, A., Linard, C., Lennert, M.,
Vanhuysse, S., Mboga, N., Wolff, E., Kalogirou, S. (2019)
Geographical Random Forests: A Spatial Extension of the Random Forest
Algorithm to Address Spatial Heterogeneity in Remote Sensing and
Population Modelling. \emph{Geocarto International}, \doi{10.1080/10106049.2019.1595177}.
}
\author{Stamatis Kalogirou \email{stamatis.science@gmail.com},
        Stefanos Georganos \email{stefanos.georganos@kau.se}}
\note{For tutorials, publications and contact information, see the
package web site.}
\seealso{\code{\link{grf}}}
\examples{
  \donttest{
    set.seed(1)
    RDF <- random.test.data(8, 8, 3)
    Coords <- RDF[, 4:5]
    m <- grf(dep ~ X1 + X2, dframe = RDF, bw = 12,
             kernel = "adaptive", coords = Coords,
             ntree = 100, mtry = 1, nthreads = 1,
             print.results = FALSE, progress = FALSE)

    set.seed(2)
    RDF.Test <- random.test.data(2, 2, 3)
    predict(m, RDF.Test, x.var.name = "X", y.var.name = "Y",
            local.w = 1, global.w = 0)
  }
  \dontrun{
    ## Real-world demonstration on the Greek Income dataset.
    ## Not run by R CMD check because fitting a full GRF on 325
    ## municipalities exceeds the few-second example budget.
    data(Income)
    Coords <- Income[, 1:2]

    m <- grf(Income01 ~ UnemrT01 + PrSect01, dframe = Income, bw = 60,
             kernel = "adaptive", coords = Coords)

    set.seed(123)
    x <- runif(20, min = 142498, max = 1001578)
    y <- runif(20, min = 3855768, max = 4606754)
    u <- runif(20, min = 5,      max = 50)
    p <- runif(20, min = 0,      max = 100)
    f <- runif(20, min = 2,      max = 30)
    df2 <- data.frame(X = x, Y = y, UnemrT01 = u, PrSect01 = p, Foreig01 = f)

    predict(m, df2, x.var.name = "X", y.var.name = "Y",
            local.w = 1, global.w = 0)
  }
}
\keyword{spatial}
\keyword{models}

\name{SpatialML-package}
\alias{SpatialML-package}
\alias{SpatialML}
\docType{package}
\title{
Spatial Machine Learning: Geographically Weighted Random Forest
}
\description{
\pkg{SpatialML} implements a spatial extension of the Random Forest
algorithm. The main functions are \code{\link{grf}} for fitting a
Geographically Weighted Random Forest (GRF), \code{\link{grf.bw}} for
selecting an optimal bandwidth, \code{\link{predict.grf}} (the S3
method dispatched by \code{predict()}) for generating spatial
predictions, and \code{\link{rf.mtry.optim}} for tuning the global
\code{mtry} parameter through cross-validation. The package also ships a
small synthetic-data generator \code{\link{random.test.data}} and a real
data set \code{\link{Income}}.
}
\details{
GRF fits one Random Forest per observation in space, using a local
neighbourhood defined by an \emph{adaptive} kernel (the \eqn{k}{k}
nearest neighbours) or a \emph{fixed} kernel (a Euclidean radius).
Observations within the neighbourhood are optionally weighted by the
bi-square kernel
\eqn{w_{ij} = (1 - (d_{ij} / h)^2)^2}, where \eqn{d_{ij}} is the
distance between observations \eqn{i} and \eqn{j} and \eqn{h} is the
local bandwidth. The approach combines the flexibility and high
predictive accuracy of Random Forests with a treatment of spatial
non-stationarity inspired by Geographically Weighted Regression.

The package uses \code{\link[ranger]{ranger}} as its random-forest
back-end. Undefined local OOB predictions are handled with a quiet
leave-one-out fallback. See
\code{vignette("SpatialML")} for a complete walk-through of a
\code{mtry} tuning -> bandwidth search -> GRF fit -> prediction
workflow.

For tutorials, related publications and contact information visit the
maintainer's website at \url{https://stamatisgeoai.eu/}.
}
\references{
Georganos, S., Grippa, T., Niang Gadiaga, A., Linard, C., Lennert, M.,
Vanhuysse, S., Mboga, N., Wolff, E., Kalogirou, S. (2019)
Geographical Random Forests: A Spatial Extension of the Random Forest
Algorithm to Address Spatial Heterogeneity in Remote Sensing and
Population Modelling. \emph{Geocarto International}, \doi{10.1080/10106049.2019.1595177}.

Georganos, S. and Kalogirou, S. (2022) A Forest of Forests: A Spatially
Weighted and Computationally Efficient Formulation of Geographical
Random Forests. \emph{ISPRS International Journal of Geo-Information},
11(9), 471. \doi{10.3390/ijgi11090471}.
}
\author{
Stamatis Kalogirou \email{stamatis.science@gmail.com} (maintainer,
\url{https://stamatisgeoai.eu/}),
Stefanos Georganos \email{stefanos.georganos@kau.se}.
}
\seealso{
\code{\link{grf}}, \code{\link{grf.bw}}, \code{\link{predict.grf}},
\code{\link{rf.mtry.optim}}, \code{\link{random.test.data}},
\code{\link{Income}}, \code{\link[ranger]{ranger}}.
}
\keyword{package}
\keyword{spatial}

\name{grf.bw}
\alias{grf.bw}
\title{Optimal Bandwidth Selection for a Geographically Weighted Random Forest}
\description{
Searches an exhaustive grid of bandwidths and returns the one that
maximises the out-of-bag (OOB) R-squared of the local model produced by
\code{\link{grf}}.
}
\usage{
grf.bw(formula, dataset, kernel = "adaptive", coords, bw.min = NULL,
       bw.max = NULL, step = 1, trees = 500, mtry = NULL,
       importance = "impurity", nthreads = 1, forests = FALSE,
       geo.weighted = TRUE, verbose = TRUE, ...)
}
\arguments{
  \item{formula}{a formula (or character string coercible to one) of the
    local model, as accepted by \code{\link[ranger]{ranger}}.}

  \item{dataset}{a data frame containing the dependent and independent
    variables that appear in \code{formula}.}

  \item{kernel}{the kernel used in the regression. Either
    \code{"adaptive"} (default) or \code{"fixed"}.}

  \item{coords}{a numeric matrix or data frame with the X and Y
    coordinates of the observations in \code{dataset}.}

  \item{bw.min}{numeric; lower bound of the bandwidth search.}

  \item{bw.max}{numeric; upper bound of the bandwidth search.}

  \item{step}{numeric; the step size of the search. Default is \code{1}.}

  \item{trees}{integer; the number of trees to grow for each local random
    forest.}

  \item{mtry}{the number of variables randomly sampled as candidates at
    each split. Default is \code{max(floor(p / 3), 1)} where \code{p} is
    the number of predictors in \code{formula}. Values above \code{p}
    are capped at \code{p}.}

  \item{importance}{variable importance measure passed to
    \code{\link[ranger]{ranger}}. Default is \code{"impurity"}.}

  \item{nthreads}{number of threads passed to
    \code{\link[ranger]{ranger}} and \code{\link[stats]{predict}}.}

  \item{forests}{logical; if \code{TRUE} all local forests are kept in
    memory. Default is \code{FALSE} (recommended for bandwidth search).}

  \item{geo.weighted}{logical; if \code{TRUE} fits a Geographically
    Weighted Random Forest using \code{ranger}'s case-weight option.}

  \item{verbose}{logical; if \code{TRUE} (default) prints one progress
    message per evaluated bandwidth and the selected bandwidth.}

  \item{\dots}{additional arguments passed to \code{\link{grf}} and
    \code{\link[ranger]{ranger}}.}
}
\details{
The function evaluates a sequence of candidate bandwidths from
\code{bw.min} to \code{bw.max} in increments of \code{step}. For each
bandwidth a Geographically Weighted Random Forest is fitted via
\code{\link{grf}} and three goodness-of-fit values are recorded:
\describe{
  \item{Local}{R-squared of the local model alone.}
  \item{Mixed}{R-squared of the equally weighted average of local and
    global predictions.}
  \item{Low.Local}{R-squared of \code{0.25 * local + 0.75 * global}.}
}
The optimal bandwidth is selected as the one that maximises the
\emph{Local} R-squared.

If you require fully reproducible bandwidth selection, call
\code{set.seed()} before \code{grf.bw()}.
}
\value{
A list with two components:
  \item{tested.bandwidths}{A data frame with one row per evaluated
    bandwidth and four columns: \code{Bandwidth}, \code{Local},
    \code{Mixed}, \code{Low.Local}.}
  \item{Best.BW}{The bandwidth that yields the highest \code{Local}
    R-squared.}
}
\references{
Georganos, S., Grippa, T., Niang Gadiaga, A., Linard, C., Lennert, M.,
Vanhuysse, S., Mboga, N., Wolff, E., Kalogirou, S. (2019)
Geographical Random Forests: A Spatial Extension of the Random Forest
Algorithm to Address Spatial Heterogeneity in Remote Sensing and
Population Modelling. \emph{Geocarto International}, \doi{10.1080/10106049.2019.1595177}.

Georganos, S. and Kalogirou, S. (2022) A Forest of Forests: A Spatially
Weighted and Computationally Efficient Formulation of Geographical
Random Forests. \emph{ISPRS International Journal of Geo-Information},
11(9), 471. \doi{10.3390/ijgi11090471}.
}
\author{Stamatis Kalogirou \email{stamatis.science@gmail.com},
        Stefanos Georganos \email{stefanos.georganos@kau.se}}
\note{Future versions may include heuristic search strategies
(e.g. \code{\link[stats]{optim}}).}
\section{Warning}{Bandwidth selection on large data sets is
computationally expensive because a full \code{\link{grf}} is fitted at
every bandwidth in the grid.}
\seealso{\code{\link{grf}}, \code{\link[ranger]{ranger}}}
\examples{
  \donttest{
    set.seed(1)
    RDF <- random.test.data(8, 8, 3)
    Coords <- RDF[, 4:5]
    bw.test <- grf.bw(dep ~ X1 + X2, dataset = RDF, kernel = "adaptive",
                      coords = Coords,
                      bw.min = 12, bw.max = 16, step = 2,
                      trees = 100, mtry = 1, nthreads = 1L,
                      forests = FALSE, geo.weighted = TRUE,
                      verbose = FALSE)
    bw.test$Best.BW
  }
  \dontrun{
    ## Real-world demonstration on the Greek Income dataset.
    ## Not run by R CMD check because an exhaustive bandwidth search on
    ## 325 municipalities is intrinsically slow.
    data(Income)
    Coords <- Income[, 1:2]

    bwe <- grf.bw(Income01 ~ UnemrT01 + PrSect01, dataset = Income,
                  kernel = "adaptive", coords = Coords,
                  bw.min = 30, bw.max = 80, step = 5,
                  forests = FALSE, geo.weighted = TRUE)

    m <- grf(Income01 ~ UnemrT01 + PrSect01, dframe = Income,
             bw = bwe$Best.BW, kernel = "adaptive", coords = Coords)
  }
}
\keyword{spatial}
\keyword{optimize}

\name{grf}
\alias{grf}
\title{Geographically Weighted Random Forest Model}
\description{
Fits a local version of the Random Forest algorithm. The function fits one
sub-model for every observation in the data set, where each local model is
trained on the geographically nearest neighbours of the focal observation
(adaptive kernel) or on every observation within a fixed distance
(fixed kernel). Local observations may further be weighted with a
bi-square kernel to reproduce a Geographically Weighted Random Forest.
}
\usage{
grf(formula, dframe, bw, kernel = c("adaptive", "fixed"), coords,
    ntree = 500, mtry = NULL, importance = "impurity", nthreads = NULL,
    forests = TRUE, geo.weighted = TRUE, print.results = TRUE,
    progress = print.results, oob.fallback = c("loo", "inbag"), ...)
}
\arguments{
  \item{formula}{a formula (or character string coercible to one) specifying
    the local model to be fitted, using the syntax of
    \code{\link[ranger]{ranger}}.}

  \item{dframe}{a data frame containing the dependent and independent
    variables that appear in \code{formula}.}

  \item{bw}{a positive number representing either the number of nearest
    neighbours (when \code{kernel = "adaptive"}) or the bandwidth in the
    units of \code{coords} (when \code{kernel = "fixed"}).}

  \item{kernel}{the type of kernel used to define the neighbourhood of each
    focal observation. One of \code{"adaptive"} (default) or
    \code{"fixed"}.}

  \item{coords}{a numeric matrix or data frame with two columns giving the
    X and Y coordinates of the observations in \code{dframe}. The number of
    rows must match \code{nrow(dframe)}.}

  \item{ntree}{integer; the number of trees to grow for each (global and
    local) random forest. Default is \code{500}.}

  \item{mtry}{the number of variables randomly sampled as candidates at
    each split. The default is \code{max(floor(p / 3), 1)} where \code{p}
    is the number of predictors in \code{formula}.}

  \item{importance}{variable importance measure. Default is
    \code{"impurity"} (see \code{\link[ranger]{ranger}}).}

  \item{nthreads}{number of threads used by \code{\link[ranger]{ranger}}
    and by \code{\link[stats]{predict}}. The default (\code{NULL}) lets
    \pkg{ranger} pick a sensible value.}

  \item{forests}{logical; if \code{TRUE} (default) all local forests are
    stored in the returned object. Required if you intend to call
    \code{\link{predict.grf}} on new data.}

  \item{geo.weighted}{logical; if \code{TRUE} (default) the local random
    forests are fitted using bi-square geographical case weights. If
    \code{FALSE}, local forests are fitted on the same neighbourhood but
    without weighting (i.e. local but unweighted).}

  \item{print.results}{logical; if \code{TRUE} (default) the function
    prints global and local model summaries to the console.}

  \item{progress}{logical; if \code{TRUE} a text progress bar is shown
    while the local random forests are fitted (only in interactive
    sessions). Defaults to \code{print.results}.}

  \item{oob.fallback}{character; fallback used when \pkg{ranger} cannot
    produce an out-of-bag prediction for the focal observation. The
    default \code{"loo"} fits one additional local forest after removing
    the focal observation and predicts it out-of-sample. \code{"inbag"}
    uses the in-sample local prediction.}

  \item{\dots}{additional arguments passed to \code{\link[ranger]{ranger}}.}
}
\details{
Geographically Weighted Random Forest (GRF) is a spatial analysis method
that fits a local version of the Random Forest algorithm to investigate
spatial non-stationarity in the relationship between a dependent variable
and a set of predictors. A sub-model is fitted for every observation in
space, considering only its neighbouring observations. The technique is
inspired by Geographically Weighted Regression (Kalogirou, 2003) but
replaces the linear local model by a non-linear ensemble that is robust
in high dimensions and less prone to over-fitting thanks to its
bootstrapping nature, which relaxes the distributional assumptions of
classical Gaussian statistics.

For each focal observation \eqn{i} the function:
\enumerate{
  \item identifies the local neighbourhood of \eqn{i} according to
    \code{kernel} and \code{bw};
  \item computes bi-square geographical weights
    \eqn{w_{ij} = (1 - (d_{ij} / h)^2)^2} where \eqn{d_{ij}} is the
    distance between observations \eqn{i} and \eqn{j} and \eqn{h} is the
    local bandwidth;
  \item fits a \pkg{ranger} random forest on the local sub-sample (with
    case weights when \code{geo.weighted = TRUE});
  \item stores the local variable importance, the OOB and in-sample
    predictions for \eqn{i}, and (optionally) the entire local forest.
}

For very small neighbourhoods or highly concentrated case weights,
\pkg{ranger} can occasionally have no out-of-bag trees for the focal
observation. In that case the OOB prediction is undefined. The default
\code{oob.fallback = "loo"} avoids noisy retry warnings by fitting one
extra local forest with the focal row removed and using that
out-of-sample prediction for the focal row.

The procedure is described in Georganos et al. (2019) and Georganos and
Kalogirou (2022).
}
\value{
A list of class \code{"grf"} with the following components:
  \item{Global.Model}{A \code{ranger} object containing the global random
    forest fitted on \code{dframe}.}
  \item{Locations}{The \code{coords} object passed to the function.}
  \item{Local.Variable.Importance}{A data frame with one row per
    observation and one column per predictor giving the local feature
    importance.}
  \item{LGofFit}{A data frame with the observed values and local
    goodness-of-fit statistics: OOB and in-sample predicted values,
    residuals, mean squared error, R-squared and the number of attempts
    needed to obtain the OOB or fallback prediction (\code{LPerm}).}
  \item{Forests}{When \code{forests = TRUE}, a list of length
    \code{nrow(dframe)} containing all local \code{ranger} forests.}
  \item{LocalModelSummary}{A list with the average local variable
    importance, mean squared error (OOB and predicted), R-squared (OOB and
    predicted), and AIC / AICc statistics.}
}
\references{
Georganos, S., Grippa, T., Niang Gadiaga, A., Linard, C., Lennert, M.,
Vanhuysse, S., Mboga, N., Wolff, E., Kalogirou, S. (2019)
Geographical Random Forests: A Spatial Extension of the Random Forest
Algorithm to Address Spatial Heterogeneity in Remote Sensing and
Population Modelling. \emph{Geocarto International}, \doi{10.1080/10106049.2019.1595177}.

Georganos, S. and Kalogirou, S. (2022) A Forest of Forests: A Spatially
Weighted and Computationally Efficient Formulation of Geographical
Random Forests. \emph{ISPRS International Journal of Geo-Information},
11(9), 471. \doi{10.3390/ijgi11090471}.

Kalogirou, S. (2003) The Statistical Analysis and Modelling of Internal
Migration Flows within England and Wales. PhD Thesis, School of
Geography, Politics and Sociology, University of Newcastle upon Tyne.
}
\author{Stamatis Kalogirou \email{stamatis.science@gmail.com},
        Stefanos Georganos \email{stefanos.georganos@kau.se}}
\note{For tutorials, publications and contact information, see the
package web site.}
\section{Warning}{Large data sets may take a long time to calibrate. A
high number of observations combined with \code{forests = TRUE} can
produce a very large output object.}
\seealso{\code{\link{predict.grf}}, \code{\link{grf.bw}},
         \code{\link[ranger]{ranger}}}
\examples{
  \donttest{
    set.seed(1)
    RDF <- random.test.data(8, 8, 3)
    Coords <- RDF[, 4:5]
    m <- grf(dep ~ X1 + X2, dframe = RDF, bw = 12,
             kernel = "adaptive", coords = Coords,
             ntree = 100, mtry = 1, nthreads = 1,
             print.results = FALSE, progress = FALSE)
  }
  \dontrun{
    ## Real-world demonstration on the Greek Income dataset.
    ## Not run by R CMD check because a full GRF on 325 municipalities
    ## with 500 trees per local fit takes longer than a CRAN example
    ## should.
    data(Income)
    Coords <- Income[, 1:2]
    m <- grf(Income01 ~ UnemrT01 + PrSect01, dframe = Income, bw = 60,
             kernel = "adaptive", coords = Coords)
  }
}
\keyword{spatial}
\keyword{models}
\keyword{regression}

\name{Income}
\alias{Income}
\docType{data}
\title{Mean Household Income at the Local Authorities of Greece in 2011}
\description{
Municipality centroids and socio-economic variables aggregated to the
new local authority geography in Greece (Programme \emph{Kallikratis}).
}
\usage{data("Income")}
\format{
A data frame with 325 observations on the following 6 variables:
  \describe{
    \item{\code{X}}{numeric; X coordinate of the municipality centroid.}
    \item{\code{Y}}{numeric; Y coordinate of the municipality centroid.}
    \item{\code{UnemrT01}}{numeric; total unemployment rate in 2001
      (Census).}
    \item{\code{PrSect01}}{numeric; proportion of the economically active
      population working in the primary sector (mainly agriculture,
      fishery and forestry) in 2001 (Census).}
    \item{\code{Foreig01}}{numeric; proportion of inhabitants without
      Greek citizenship in 2001 (Census).}
    \item{\code{Income01}}{numeric; mean household income, in Euros,
      earned in 2001 and declared in 2002 tax forms.}
  }
}
\details{
The X, Y coordinates refer to the geometric centroids of the 325
municipalities in Greece established by the \emph{Kallikratis} reform in
2011.
}
\source{
The original shapefile of the corresponding polygons was published by
the Hellenic Statistical Authority (EL.STAT.) at
\url{https://www.statistics.gr/}. The population, employment, citizenship
and employment-sector data were also published by EL.STAT.; the income
data were obtained from the General Secretariat of Information Systems
in Greece at the postcode level. All data were aggregated to the new
municipalities by the author.
}
\references{
Kalogirou, S. and Hatzichristos, T. (2007) A spatial modelling framework
for income estimation. \emph{Spatial Economic Analysis}, 2(3), 297-316.
\doi{10.1080/17421770701576921}.

Kalogirou, S. (2010) Spatial inequalities in income and post-graduate
educational attainment in Greece. \emph{Journal of Maps}, 6(1), 393-400.
\doi{10.4113/jom.2010.1095}.
}
\examples{
data(Income)
boxplot(Income$Income01)
hist(Income$PrSect01)
}
\keyword{datasets}

\name{rf.mtry.optim}
\alias{rf.mtry.optim}
\title{Optimal mtry for a Random Forest via OOB or Cross-Validation}
\description{
Searches for the value of \code{mtry} that minimises the predictive error
of a Random Forest on a user-supplied grid. Three evaluation strategies
are available:
\itemize{
  \item \code{"oob"} (default): one \code{\link[ranger]{ranger}} fit per
    \code{mtry}; OOB error is the criterion (fast).
  \item \code{"cv"}: a single \code{cv.folds}-fold cross-validation is
    performed for each \code{mtry}.
  \item \code{"repeatedcv"}: \code{cv.repeats} repeats of \code{cv.folds}-fold
    CV are performed for each \code{mtry}.
}
The selected value can then be passed to the \code{mtry} argument of
\code{\link{grf}}.
}
\usage{
rf.mtry.optim(formula, dataset, min.mtry = NULL, max.mtry = NULL,
              mtry.step = 1, num.trees = 500,
              cv.method = c("oob", "repeatedcv", "cv"),
              cv.folds = 10, cv.repeats = 5, num.threads = NULL,
              plot.it = TRUE, verbose = TRUE, ...)
}
\arguments{
  \item{formula}{a model formula (or a character string coercible to
    one).}
  \item{dataset}{a data frame containing the variables of \code{formula}.}
  \item{min.mtry}{integer; lower bound of the \code{mtry} grid. Default
    \code{1}.}
  \item{max.mtry}{integer; upper bound of the \code{mtry} grid. Default
    is the number of predictors in \code{formula}.}
  \item{mtry.step}{integer; step of the \code{mtry} grid. Default
    \code{1}.}
  \item{num.trees}{integer; number of trees in each random forest fit.
    Default \code{500}.}
  \item{cv.method}{character; evaluation strategy. One of \code{"oob"}
    (default), \code{"repeatedcv"} or \code{"cv"}.}
  \item{cv.folds}{integer; number of folds when
    \code{cv.method != "oob"}. Default \code{10}.}
  \item{cv.repeats}{integer; number of repeats when
    \code{cv.method == "repeatedcv"}. Default \code{5}.}
  \item{num.threads}{number of threads passed to
    \code{\link[ranger]{ranger}}. The default (\code{NULL}) lets
    \pkg{ranger} pick a sensible value.}
  \item{plot.it}{logical; if \code{TRUE} (default) the \code{mtry} vs.
    RMSE curve is plotted.}
  \item{verbose}{logical; if \code{TRUE} (default) progress messages are
    printed for every evaluated \code{mtry}.}
  \item{\dots}{additional arguments forwarded to
    \code{\link[ranger]{ranger}} (e.g. \code{importance = "impurity"}).}
}
\details{
The criterion to minimise is the average RMSE across folds (or the OOB
RMSE in the \code{"oob"} case). The average squared correlation between
predictions and observations is also recorded as the \code{Rsquared}
column.

For reproducible results call \code{set.seed()} before
\code{rf.mtry.optim()}.

\strong{Compatibility note:} the function no longer returns a
\code{caret} \code{train} object. The dependency on \pkg{caret} (and
indirectly on \pkg{randomForest}) has been removed and the
cross-validation is now implemented directly on top of
\code{\link[ranger]{ranger}}. Code that read \code{$bestTune$mtry} from
the previous return value should now read \code{$best.mtry}.
}
\value{
A list with the following components:
  \item{best.mtry}{The \code{mtry} value with the lowest average RMSE.}
  \item{results}{A data frame with one row per evaluated \code{mtry} and
    the columns \code{mtry}, \code{RMSE}, \code{Rsquared}, \code{SDRMSE}
    and \code{SDRsq}. Standard deviations are \code{NA} for
    \code{cv.method == "oob"}.}
  \item{cv.method}{The evaluation strategy used.}
  \item{num.trees}{Number of trees used in each fit.}
  \item{call}{The matched call.}
}
\references{
Wright, M. N. and Ziegler, A. (2017) ranger: A Fast Implementation of
Random Forests for High Dimensional Data in C++ and R.
\emph{Journal of Statistical Software}, 77(1), 1-17.
\doi{10.18637/jss.v077.i01}.

Georganos, S. and Kalogirou, S. (2022) A Forest of Forests: A Spatially
Weighted and Computationally Efficient Formulation of Geographical
Random Forests. \emph{ISPRS International Journal of Geo-Information},
11(9), 471. \doi{10.3390/ijgi11090471}.
}
\author{Stamatis Kalogirou \email{stamatis.science@gmail.com},
        Stefanos Georganos \email{stefanos.georganos@kau.se}}
\seealso{\code{\link{grf}}, \code{\link[ranger]{ranger}}}
\examples{
  \donttest{
    data(Income)
    set.seed(123)
    res <- rf.mtry.optim(Income01 ~ UnemrT01 + PrSect01, dataset = Income,
                         num.trees = 200, num.threads = 1L,
                         cv.method = "oob", plot.it = FALSE,
                         verbose = FALSE)
    res$best.mtry
    res$results
  }
}
\keyword{models}
\keyword{optimize}

\name{random.test.data}
\alias{random.test.data}
\title{Random Data Generator}
\description{
Generates a synthetic data set composed of one dependent variable, a
number of independent variables drawn from \code{Uniform(0, 1)}, and
optionally the X and Y coordinates of a regular \code{nrows} x
\code{ncols} grid. Useful for examples, unit testing and small
simulation studies.
}
\usage{
random.test.data(nrows = 10, ncols = 10, vars.no = 3,
                 dep.var.dis = c("normal", "poisson"),
                 xycoords = TRUE)
}
\arguments{
  \item{nrows}{integer; number of rows of the regular grid. Default
    \code{10}.}
  \item{ncols}{integer; number of columns of the regular grid. Default
    \code{10}.}
  \item{vars.no}{integer; total number of model variables (1 dependent
    plus \code{vars.no - 1} independent). Must be at least \code{2}.}
  \item{dep.var.dis}{distribution of the dependent variable. One of
    \code{"normal"} (default, drawn from \code{rnorm()}) or
    \code{"poisson"} (drawn from \code{rpois()} with \code{lambda = 7}).}
  \item{xycoords}{logical; if \code{TRUE} (default) two columns \code{X}
    and \code{Y} containing the grid coordinates are appended.}
}
\details{
The function is mostly intended to provide reproducible toy data for
the package examples. The independent variables are named
\code{X1, X2, \dots, X(vars.no - 1)}.

For reproducible results call \code{set.seed()} before
\code{random.test.data()}.
}
\value{
A data frame with \code{nrows * ncols} rows containing one column
\code{dep}, \code{vars.no - 1} columns named \code{X1, X2, ...} and,
when \code{xycoords = TRUE}, two extra columns \code{X} and \code{Y}.
}
\author{Stamatis Kalogirou \email{stamatis.science@gmail.com}}
\examples{
set.seed(1)
RDF <- random.test.data(12, 12, 3)
head(RDF)
}
\keyword{datagen}
